/* wptGPGMEWrapper.cpp - Object oriented wrapper for GPGME
 *	Copyright (C) 2006, 2007 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <stdio.h>
#include <string.h>

#include "gpgme.h"
#include "wptGPG.h"
#include "wptGPGME.h"
#include "wptW32API.h"


/* Constructor to build gpgme context. */
GPGME::GPGME ()
{
    gpgme_new (&ctx);
}

/* Destructor to release gpgme context. */
GPGME::~GPGME ()
{
    if (ctx != NULL)
	gpgme_release (ctx);
}

/* Enable or disable the feature to add the 
    WinPT version to armored output. */
void
GPGME::setChangeVersion (bool val)
{
    chg_ver = val;
}

/* Enable or disable armor. */
void 
GPGME::setArmor (bool val)
{
    gpgme_set_armor (ctx, val? 1 : 0);
}

bool
GPGME::getArmor (void)
{
    return gpgme_get_armor (ctx)? true: false;
}


/* Import keys from a memory buffer. 
   Return value: 0 on success. */
gpgme_error_t
GPGME::importFromBuffer (const char *indata)
{
    gpgme_data_t keydata = NULL;
    gpgme_error_t err;

    err = gpgme_data_new_from_mem (&keydata, indata, strlen (indata), 1);
    if (err)
	return err;

    op_begin ();
    err = gpgme_op_import (ctx, keydata);
    op_end ();

    gpgme_data_release (keydata);
    return err;
}


/* Import the keys from the clipboard.
   Return value: 0 on success. */
gpgme_error_t
GPGME::importFromClipboard (void)
{    
    gpgme_data_t keydata = NULL;
    gpgme_error_t err;

    err = gpg_data_new_from_clipboard (&keydata, 0);
    if (err)
	return err;

    op_begin ();
    err = gpgme_op_import (ctx, keydata);
    op_end ();

    gpgme_data_release (keydata);
    return err;
}


/* Import the keys from the file @file.
   Return value: 0 on success. */
gpgme_error_t
GPGME::importFromFile (const char *file)
{
    gpgme_data_t keydata;
    gpgme_error_t err;

    err = gpgme_data_new_from_file (&keydata, file, 1);
    if (err)
	return err;
    
    op_begin ();
    err = gpgme_op_import (ctx, keydata);
    op_end ();

    gpgme_data_release (keydata);
    return err;
}


/* Return the import result for the last GPG operation. */
gpgme_import_result_t
GPGME::importGetResult (void)
{
    return gpgme_op_import_result (ctx);
}


/* Export a single key or all keys which matches the pattern @patt 
   to file @outfile. 
   Return value: 0 on success. */
gpgme_error_t
GPGME::exportToFile (const char *patt, const char *outfile)
{
    gpgme_data_t dat;
    gpgme_error_t err;

    err = gpgme_data_new (&dat);
    if (err)
	return err;
    op_begin ();
    err = gpgme_op_export (ctx, patt, 0, dat);
    op_end ();
    if (err)
	gpgme_data_release (dat);
    else {
	err = gpg_data_release_and_set_file (dat, outfile);
	if (err)
	    gpgme_data_release (dat);
    }
    
    return err;
}

/* Export multiple keys using @pattern. */
gpgme_error_t
GPGME::exportToFile (const char *pattern[], const char *outfile)
{
    gpgme_error_t err;
    gpgme_data_t dat;
    
    err = gpgme_data_new (&dat);
    if (err)
	return err;
    
    op_begin ();
    err = gpgme_op_export_ext (ctx, pattern, 0, dat);
    op_end ();
    if (err)
	gpgme_data_release (dat);
    else {
	err = gpg_data_release_and_set_file (dat, outfile);
	if (err)
	    gpgme_data_release (dat);
    }
    
    return err;
}
/* Export key pattern @patt to the clipboard.
   Return value: 0 on success. */
gpgme_error_t
GPGME::exportToClipboard (const char *patt)
{
    gpgme_data_t dat;
    gpgme_error_t err;

    err = gpgme_data_new (&dat);
    if (err)
	return err;
    op_begin ();
    err = gpgme_op_export (ctx, patt, 0, dat);
    op_end ();
    if (err)
	gpgme_data_release (dat);
    else
	gpg_data_release_to_clipboard (dat, chg_ver? 1 : 0);
    return err;
}

gpgme_error_t
GPGME::exportToClipboard (const char *pattern[])
{
    gpgme_data_t dat;
    gpgme_error_t err;
    
    err = gpgme_data_new (&dat);
    if (err)
	return err;
    op_begin ();
    err = gpgme_op_export_ext (ctx, pattern, 0, dat);
    op_end ();
    if (err)
	gpgme_data_release (dat);
    else
	gpg_data_release_to_clipboard (dat, chg_ver? 1: 0);
    return err;
}

/* Export key pattern @patt to the buffer @outdata. 
   Return value: 0 on success. */
gpgme_error_t
GPGME::exportToBuffer (const char *patt, char **outdata)
{
    gpgme_error_t err;
    gpgme_data_t dat;
    size_t n;
    char *p;

    err = gpgme_data_new (&dat);
    if (err)
	return err;
    op_begin ();
    err = gpgme_op_export (ctx, patt, 0, dat);
    op_end ();
    if (err) {
	gpgme_data_release (dat);
	return err;
    }
    p = gpgme_data_release_and_get_mem (dat, &n);

    /* we use this detour because p were allocated inside
       the gpgme DLL and cannot be freed with 'free' because
       the DLL might be linked against debug libs but the application
       run ins release mode. */
    *outdata = new char[n+1];
    memcpy (*outdata, p, n);
    (*outdata)[n] = 0;
    gpgme_free (p);

    return err;
}
