/* wptKeyserverDlg.cpp - Keyserver dialog
 *	Copyright (C) 2000-2007, 2009 Timo Schulz
 *	Copyright (C) 2005, 2006 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <commctrl.h>
#include <ctype.h>

#include "resource.h"
#include "wptErrors.h"
#include "wptTypes.h"
#include "wptCommonCtl.h"
#include "wptNLS.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptGPG.h"
#include "wptGPGME.h"
#include "wptKeyserver.h"
#include "wptContext.h" /* for passphrase_s */
#include "wptKeyManager.h"
#include "wptDlgs.h"
#include "wptUTF8.h"
#include "wptRegistry.h"
#include "wptKeylist.h"


/* Symbolic column IDs */
enum { KS_COL_NAME = 0, KS_COL_PORT };


/* Print out keyserver error and a possible Winsock error. */
static void
hkp_err_box (HWND dlg, const char *host, WORD port, int rc)
{
    const char *err = kserver_strerror ();
    const char *ws_err = wsock_strerror ();
    
    if (ws_err && strlen (ws_err) > 0)
	log_box (_("Keyserver"), MB_ERR, "%s:%d: %s", host, port, ws_err);
    else
	log_box (_("Keyserver"), MB_ERR, "%s:%d: %s\n%s",
		 host, port, winpt_strerror (rc), err? err : "");
}


/* Send the extracted key with keyid @pattern to the 
   keyserver @kserver (port @port).
   Return value: 0 on success. */
int
hkp_send_key (HWND dlg, const char *kserver, WORD port, const char *pattern)
{
    GPGME gpg;
    gpgme_error_t ec;
    char *rawkey = NULL;
    char msg[256];
    
    gpg.setArmor (true);
    ec = gpg.exportToBuffer (pattern, &rawkey);
    if (ec) {
        msg_box (dlg, gpgme_strerror (ec), _("Export"), MB_ERR);
        goto leave;
    }
    op_begin ();
    ec = kserver_sendkey (kserver, port, rawkey, strlen (rawkey));
    op_end ();
    if (ec) {
        hkp_err_box (dlg, kserver, port, ec);
        goto leave;
    }
    
    _snprintf (msg, DIM (msg) -1, _("Key '%s' successfully sent"), pattern);
    status_box (dlg, msg, _("GnuPG Status"));
    
leave:
    free_if_alloc (rawkey);
    return ec? -1 : 0;
}



/* Show all received keys from the keyserver. */
static void
show_imported_keys (gpgme_import_result_t res)
{
    gpgme_import_status_t t;
    gpgme_key_t key;
    gpgme_data_t msg;
    gpgme_ctx_t ctx;
    gpgme_error_t err;
    const char *s;
    char *p, *uid;
    size_t n=0;

    if (!res)
	return;

    err = gpgme_data_new (&msg);
    if (err)
	BUG (NULL);
    err = gpgme_new (&ctx);
    if (err)
	BUG (NULL);
    if (res->considered > 1) {
	s = _("WARNING: multiple keys matched request.\n\n");
	gpgme_data_write (msg, s, strlen (s));
    }

    if (res->unchanged == res->considered)
	s = _("Key(s) successfully received but nothing was changed.");
    else
	s = _("Key(s) sucessfully received and imported.");
    // nice to have the amount of new signatures...
    gpgme_data_write (msg, s, strlen (s));
    gpgme_data_write (msg, "\n\n", 2);

    for (t=res->imports; t; t = t->next) {
	if (!gpgme_get_key (ctx, t->fpr, &key, 0)) {
	    s = uid = utf8_to_native (key->uids->uid);
	    gpgme_data_write (msg, s, strlen (s));
	    gpgme_data_write (msg, "\n", 1);
	    gpgme_key_release (key);
	    safe_free (uid);
	}
    }
    gpgme_data_write (msg, "\0", 1);
    p = gpgme_data_release_and_get_mem (msg, &n);
    if (p != NULL) {
	msg_box (NULL, p, _("Imported Keys"), MB_INFO);
	gpgme_free (p);
    }
    gpgme_release (ctx);
}


/* Receive a key from the keyserver @kserver (port @port)
   with the pattern @pattern.
   Return value: 0 on success. */
static int
keyserver_recv_key (HWND dlg, const char *kserver, WORD port,
		    const char *pattern, int proto, int flags,
		    char **r_fpr)
{    
    gpgme_import_result_t import_res = NULL;
    gpgme_error_t err;
    GPGME gpg;
    char *rawkey = NULL;
    size_t keylen = 0;
    int rc;

    if ((rc = kserver_recvkey (kserver, port, 
			       kserver_check_keyid (pattern),
			       &rawkey, &keylen))) {
        hkp_err_box (dlg, kserver, port, rc);
	free_if_alloc (rawkey);
        return rc;
    }

    if (!strstr (rawkey, "BEGIN PGP PUBLIC KEY BLOCK")) {
        msg_box (dlg, _("This is not a valid OpenPGP key."),
		 _("Keyserver"), MB_ERR);
	goto leave;
    }   
    err = gpg.importFromBuffer (rawkey);
    if (err) {
        msg_box (dlg, gpgme_strerror (err), _("Import"), MB_ERR);
	goto leave;
    }
    import_res = gpg.importGetResult ();
    if (import_res && r_fpr)
	*r_fpr = m_strdup (import_res->imports->fpr);
    
    /* if we use the refresh mode, a lot of keys will be fetched and thus only
       a summarize at the end is presented and not for each key. */
    if (import_res && !(flags & KM_KS_REFRESH)) {
	show_imported_keys (import_res);
	if (import_res->unchanged == import_res->considered) {
	    rc = WPTERR_GENERAL; /* no keys updated. */
	    goto leave;
	}
    }

leave:
    free_if_alloc (rawkey);
    return rc;
}


/* Alias for keyserver_recv_key but without the ability to return 
   the fingerprint of the key. */
int
hkp_recv_key (HWND dlg, const char *kserver, WORD port,
	      const char *pattern, int proto, int flags)
{
    return keyserver_recv_key (dlg, kserver, port, pattern, 
			       proto, flags, NULL);
}

int
hkp_recv_key2 (HWND dlg, const char *kserver, WORD port,
	       const char *pattern, int proto, char **r_fpr)
{
    return keyserver_recv_key (dlg, kserver, port, pattern,
			       proto, 0, r_fpr);
}


/* Utility function to fetch a key from the keyserver
   based on a given signature (@sig).
   Return 0 on success. */
int
fetch_key_from_keyserver (HWND dlg, gpgme_signature_t sig)
{
    const char *keyid;
    char timebuf[128];
    int id;
    
    // XXX: we have a problem with v3 keys here
    if (!sig->fpr)
	return FALSE;
    if (!get_locale_timedate (sig->timestamp, timebuf, DIM (timebuf)-1))
	_snprintf (timebuf, DIM (timebuf)-1, "'unknown time'");
    keyid = get_keyid_from_fpr (sig->fpr);	    
    id = log_box (_("Verify"), MB_INFO|MB_YESNO,
		  _("Signature made %s using %s key ID 0x%s\n"
		    "Cannot check signature: public key not found\n\n"
		    "Do you want to try to retrieve the key from the keyserver?"),
		  timebuf, get_key_pubalgo (sig->pubkey_algo), keyid);
    if (id == IDNO) {
	msg_box (dlg, get_gpg_sigstat (GPGME_SIGSUM_KEY_MISSING), 
		 _("Verify"), MB_WARN);
        return -1;
    }

    if (!hkp_recv_key (dlg, default_keyserver, default_keyserver_port, 
		       keyid, 0, 0)) {
	keycache_update (0, keyid);
	return 0;
    }
    return -1;
}


/* Check if the given pattern are either a valid 
   email address, a {long, short} keyid or a fingerprint. 
   Return 0 on success. */
static int
check_pattern (const char *pattern)
{    
    size_t i;

    if (strchr (pattern, ' '))
	return WPTERR_GENERAL; /* do not allow white spaces. */
    
    if (strchr (pattern, '@') && strlen (pattern) >= 3 &&
	!check_email_address (pattern))
        return 0;

    if (strstr (pattern, "0x"))
	pattern += 2;
    if (strlen (pattern) != 8 &&
	strlen (pattern) != 16 &&
	strlen (pattern) != 40)
	return WPTERR_GENERAL;

    for (i=0; i < strlen (pattern); i++) {
	if (!isxdigit (pattern[i]))
	    return WPTERR_GENERAL;
    }
    return 0;
}


/* Return human readable name for the proxy protocol. */
static const char*
name_from_proto (int proto)
{
    const char *s;

    switch (proto) {
    case PROXY_PROTO_NONE:	s = ""; break;
    case PROXY_PROTO_HTTP:	s = "HTTP"; break;
    case PROXY_PROTO_SOCKS5:	s = "SOCKS5"; break;
    default:			s = "HTTP"; break;
    }
    return s;
}


static int inline
kserver_get_pos (listview_ctrl_t lv)
{
    if (listview_count_items (lv, 0) == 1)
	return 0;
    return listview_get_curr_pos (lv);
}


static WORD inline
kserver_get_port (listview_ctrl_t lv)
{
    char buf[16];

    listview_get_item_text (lv, kserver_get_pos (lv), 
			    KS_COL_PORT, buf, DIM (buf)-1);
    return (WORD)strtoul (buf, NULL, 10);
}


int
keyserver_list_build (listview_ctrl_t *r_lv, HWND hwnd)
{
    struct listview_column_s keyserver[] = {
        {0, 180, (char *)_("DNS Name")},
	{1,  55, (char *)_("Port")},
        {0,   0, NULL}
    };
    HICON ico[1];
    listview_ctrl_t lv;
    char buf[32];

    ico[0] = LoadIcon (glob_hinst, (LPCTSTR)IDI_COMPUTER);
    listview_new (&lv, hwnd);    
    for (int j=0; keyserver[j].fieldname; j++)
        listview_add_column (lv, &keyserver[j]);
    listview_set_image_list (lv, 16, 16, ico, 1);
    for (int j = 0; j < MAX_KEYSERVERS; j++) {
	if (!server[j].used)
	    continue;
	listview_add_item_image (lv, " ", 0);
	listview_add_sub_item (lv, 0, KS_COL_NAME, server[j].name);
	sprintf (buf, "%d", server[j].port);
	listview_add_sub_item (lv, 0, KS_COL_PORT, buf);
    }
    listview_set_ext_style (lv);
    *r_lv = lv;
    return 0;
}


static void
fill_keyserv_types (HWND dlg)
{
    HWND h = GetDlgItem (dlg, IDC_KSERVADD_TYPES);
    combox_add_string (h, _("HKP Keyserver"));
    SendMessage (h, CB_SETCURSEL, 0, 0);
}


/* XXX: factor out the pattern code for generic use. */

/* Add search pattern @patt to the combo box if it is
   not already available in the list. */
static void
add_pattern_to_combox (HWND dlg, int ctlid, const char *patt)
{
    int err;

    err = SendDlgItemMessage (dlg, ctlid, CB_FINDSTRINGEXACT,
			      0, (LPARAM)(LPCSTR)patt);
    if (err != CB_ERR)
	return;
    SendDlgItemMessage (dlg, IDC_KEYSERVER_SEARCH, CB_ADDSTRING,
			0, (LPARAM)(LPCSTR)patt);
}


/* Restore saved pattern from the registry back to the combo box.
   If no pattern were saved, do nothing and just return. */
static void
load_pattern_to_combox (HWND dlg, int ctlid, const char *rkey_name)
{
    char *p, *tok;

    p = get_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", rkey_name);
    if (!p || strlen (p) < 2) {
	free_if_alloc (p);
	return;
    }
    tok = strtok (p, "$");
    while (tok != NULL) {
	SendDlgItemMessage (dlg, ctlid, CB_ADDSTRING, 0, (LPARAM)(LPCTSTR)tok);
	tok = strtok (NULL, "$");
    }
    free_if_alloc (p);
}


/* Save all entries from the combo box given in @ctlid to the
   registry. Concat all values and use '$' as a separator. */
static void
save_pattern_from_combox (HWND dlg, int ctlid, const char *rkey_name)
{
    char *p, tmp[64];
    int i;
    int n, len=0;

    n = SendDlgItemMessage (dlg, ctlid, CB_GETCOUNT, 0, 0);
    if (n == 0 || n == CB_ERR)
	return;
    /* it is very unlikely that the combox contain hundred of pattern 
       but even so we limit the value to 16 which should be sufficient. */
    if (n > 16) n = 16;
    for (i=0; i < n; i++)
	len += SendDlgItemMessage (dlg, ctlid, CB_GETLBTEXTLEN, 
				   (WPARAM)i, 0) + 1 + 1;
    p = new char[len+1];
    memset (p, 0, len+1);
    for (i=0; i < n; i++) {
	memset (tmp, 0, sizeof (tmp));
	SendDlgItemMessage (dlg, ctlid, CB_GETLBTEXT,
			    (WPARAM)i, (LPARAM)tmp);
	strcat (p, tmp);
	strcat (p, "$");
    }
    set_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", rkey_name, p);
    free_if_alloc (p);
}


/* Dialog box procedure to access keyservers. */
BOOL CALLBACK
keyserver_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static listview_ctrl_t lv = NULL;
    static int lv_idx = 0;
    keyserver_ctx ksc;
    char kserver[128], pattern[128];
    char proto[16], *fpr = NULL;
    int rc, proto_nr = 0;
    
    switch (msg) {
    case WM_INITDIALOG:
        SetWindowText (dlg, _("Keyserver Access"));
	SetDlgItemText (dlg, IDC_KEYSERVER_RECV, _("&Receive"));
        SetDlgItemText (dlg, IDC_KEYSERVER_SEND, 
                        _("Send key (default is receiving)"));
        SetDlgItemText (dlg, IDC_KEYSERVER_INFO,
                        _("Key ID or email address you want to search for"));
        SetDlgItemText (dlg, IDC_KEYSERVER_INDEX, _("&Search"));
	SetDlgItemText (dlg, IDC_KEYSERVER_PROXSETT, _("C&hange proxy"));
	SetDlgItemText (dlg, IDC_KEYSERVER_DEFAULT, _("Set &default"));
	SetDlgItemText (dlg, IDCANCEL, _("&Close"));
	
	load_pattern_to_combox (dlg, IDC_KEYSERVER_SEARCH, "KSsearch");
        keyserver_list_build (&lv, GetDlgItem (dlg, IDC_KEYSERVER_LIST));
	center_window (dlg, NULL);
        SetForegroundWindow (dlg);
        return TRUE;
        
    case WM_NOTIFY:
        NMHDR *notify;
        notify = (NMHDR *)lparam;
	if (!notify)
	    break;
        if (notify->code == (UINT)NM_CLICK
            && notify->idFrom == (UINT)IDC_KEYSERVER_LIST)
            lv_idx = listview_get_curr_pos (lv);	
        return TRUE;
        
    case WM_DESTROY:
        if (lv) {
            listview_release (lv);
            lv = NULL;
        }
        lv_idx = 0;
	save_pattern_from_combox (dlg, IDC_KEYSERVER_SEARCH, "KSsearch");
        return FALSE;
        
    case WM_SYSCOMMAND:
        if (LOWORD (wparam) == SC_CLOSE)
            EndDialog (dlg, TRUE);
        return FALSE;	

    case WM_COMMAND:
        switch (LOWORD (wparam)) {
        case IDC_KEYSERVER_PROXSETT:
            dialog_box_param (glob_hinst, (LPCTSTR)IDD_WINPT_KEYSERVER_PROXY, 
			      dlg, keyserver_proxy_dlg_proc, 0,
                              _("Proxy Settings"), IDS_WINPT_KEYSERVER_PROXY);
            return TRUE;
            
        case IDC_KEYSERVER_INDEX:
            if (!lv_idx)
                lv_idx = kserver_get_pos (lv);
            
            if (!GetDlgItemText (dlg, IDC_KEYSERVER_SEARCH, 
				 pattern, DIM (pattern)-1)) {
                msg_box (dlg, _("Please enter the search pattern."), 
			 _("Keyserver"), MB_INFO);
                return TRUE;
            }

	    if (lv_idx != -1) {
		listview_get_item_text (lv, lv_idx, KS_COL_NAME, 
					kserver, DIM (kserver)-1);
		ksc.name = kserver;
		ksc.port = kserver_get_port (lv);
	    }
	    else {
		ksc.name = DEF_HKP_KEYSERVER;
		ksc.port = HKP_PORT;
	    }
            ksc.pattern = pattern;
            DialogBoxParam (glob_hinst, (LPCSTR)IDD_WINPT_HKPSEARCH, dlg,
                            hkpsearch_dlg_proc, (LPARAM) &ksc);
            return TRUE;
            
        case IDC_KEYSERVER_RECV:	    
            memset (&kserver, 0, sizeof (kserver));
            if (!lv_idx) {
                lv_idx = kserver_get_pos (lv);
                if (lv_idx == -1) {
                    msg_box (dlg, _("Please select one of the keyservers."), 
			    _("Keyserver"), MB_INFO);
                    return TRUE;
                }
            }
            listview_get_item_text (lv, lv_idx, KS_COL_NAME, 
				    proto, DIM (proto)-1);
	    proto_nr = KSPROTO_HTTP;
            listview_get_item_text (lv, lv_idx, KS_COL_NAME, 
				    kserver, DIM (kserver)-1);
            if (!GetDlgItemText(dlg, IDC_KEYSERVER_SEARCH,
				pattern, DIM (pattern)-1)) {
                msg_box (dlg, _("Please enter the search pattern."),
			 _("Keyserver"), MB_INFO);
                return FALSE;
            }
            if (check_pattern (pattern)) {
                msg_box (dlg, _("Only email addresses or keyids are allowed."),
			 _("Keyserver"), MB_INFO);
                return TRUE;
            }
	    rc = keyserver_recv_key (dlg, kserver, kserver_get_port (lv),
				     pattern, proto_nr, 0, &fpr);				     
	    if (!rc && fpr != NULL) {
		keycache_update (0, fpr);
		free_if_alloc (fpr);
	    }
	    if (!rc)
		add_pattern_to_combox (dlg, IDC_KEYSERVER_SEARCH, pattern);
            return TRUE;

        case IDCANCEL:
            EndDialog (dlg, FALSE);
            return TRUE;
        }
        break;
    }
    
    return FALSE;
}
